//
//  ES1Renderer.m
//  Snake
//
//  Created by Vahid Kazemi on 7/12/10.
//  Copyright KTH 2010. All rights reserved.
//

#import "ES1Renderer.h"
#import "World.h"

@implementation ES1Renderer

// Create an OpenGL ES 1.1 context
- (id)init
{
    if ((self = [super init]))
    {
        context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES1];

        if (!context || ![EAGLContext setCurrentContext:context])
        {
            [self release];
            return nil;
        }

        // Create default framebuffer object. The backing will be allocated for the current layer in -resizeFromLayer
        glGenFramebuffersOES(1, &defaultFramebuffer);
        glGenRenderbuffersOES(1, &colorRenderbuffer);
        glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
        glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
        glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_RENDERBUFFER_OES, colorRenderbuffer);
		
		texture = [self loadTexture:@"block.png"];
		accel.x = accel.y = accel.z = 1;
    }

    return self;
}

- (GLuint)loadTexture:(NSString *)filename
{
	UIImage *image = [UIImage imageNamed:filename];
    if (image == nil){
        NSLog(@"Do real error checking here");
		return 0;
	}
	
    GLuint width = CGImageGetWidth(image.CGImage);
    GLuint height = CGImageGetHeight(image.CGImage);
    void *imageData = malloc( height * width * 4 );
    CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
	CGContextRef bitmapContext = CGBitmapContextCreate( imageData, width, height, 8, 4 * width, colorSpace, kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big );
	CGColorSpaceRelease( colorSpace );
	CGContextClearRect( bitmapContext, CGRectMake( 0, 0, width, height ) );
	CGContextTranslateCTM( bitmapContext, 0, height - height );
	CGContextDrawImage( bitmapContext, CGRectMake( 0, 0, width, height ), image.CGImage );
	
	GLuint texid;
	glGenTextures(1, &texid);
	glBindTexture(GL_TEXTURE_2D, texid);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, imageData);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR); 
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	
    CGContextRelease(bitmapContext);
	
    free(imageData);
	
	return texid;
}

- (void)accelerate:(UIAcceleration*)acc 
{
	if(fabsf(acc.x)>1.5f || fabsf(acc.y)>1.5f || fabsf(acc.z)>1.5){
		accel.x = 10*acc.x;
		accel.y = 10*acc.y;
		accel.z = 10*acc.z;
	}
	accel.x = accel.x*0.9 + 0.1;
	accel.y = accel.y*0.9 + 0.1;
	accel.z = accel.z*0.9 + 0.1;
}

- (void)render
{
	float blockWidth = 2.0f/kBoardWidth;
	float blockHeight = 2.0f/kBoardHeight;
	
    GLfloat squareVertices[] = {
        0,			0,
        blockWidth,	0,
        0,			blockHeight,
        blockWidth,	blockHeight
    };
	
	GLfloat texcoords[] = {
		0, 0,
		1, 0,
		0, 1,
		1, 1,
	};

    // This application only creates a single context which is already set current at this point.
    // This call is redundant, but needed if dealing with multiple contexts.
    [EAGLContext setCurrentContext:context];

    // This application only creates a single default framebuffer which is already bound at this point.
    // This call is redundant, but needed if dealing with multiple framebuffers.
    glBindFramebufferOES(GL_FRAMEBUFFER_OES, defaultFramebuffer);
    glViewport(0, 0, backingWidth, backingHeight);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
	
	static int lastColorCount = 0;
	static float lastColor[] = {0, 0, 0};
	static float newColor[] = {0, 0, 0};
	if(lastColorCount--<=0){
		float sum = 0.0001f;
		for(int i=0; i<3; i++){
			newColor[i] = (float)rand()/RAND_MAX;
			sum += newColor[i];
		}
		float factor = 0.5f*(float)rand()/RAND_MAX/sum;
		for(int i=0; i<3; i++){
			newColor[i] *= factor;
		}
		lastColorCount = 100;
	}
	for(int i=0; i<3; i++){
		lastColor[i] = lastColor[i]*0.99f + newColor[i]*0.01f;
	}

    glClearColor(lastColor[0], lastColor[1], lastColor[2], 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);

    glVertexPointer(2, GL_FLOAT, 0, squareVertices);
    glEnableClientState(GL_VERTEX_ARRAY);
	glTexCoordPointer(2, GL_FLOAT, 0, texcoords);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
	
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture);

	static Vector2f lastRandom[kBoardWidth*kBoardHeight] = {0};
	static Vector2f newRandom[kBoardWidth*kBoardHeight] = {0};
	static int lastRandCount[kBoardWidth*kBoardHeight] = {0};
	
	// Draw board
	BlockType* board = [[World getInstance] getBoard];
	for(int j=0; j<kBoardHeight; j++){
		for(int i=0; i<kBoardWidth; i++){	
			int idx = i+j*kBoardWidth;
			
			// create a random number		
			if(lastRandCount[idx]<=0){
				newRandom[idx].x = 0.02f*((float)rand()*2/RAND_MAX-1)*accel.x;
				newRandom[idx].y = 0.02f*((float)rand()*2/RAND_MAX-1)*accel.y;
				lastRandCount[idx] = 20;
			}
			else{
				lastRandCount[idx]--;
			}
			lastRandom[idx].x = newRandom[idx].x*0.1f+lastRandom[idx].x*0.9f;
			lastRandom[idx].y = newRandom[idx].y*0.1f+lastRandom[idx].y*0.9f;
			
			float x = i*blockWidth-1+lastRandom[idx].x;
			float y = 1.0f-(j+1)*blockHeight+lastRandom[idx].y;
			glLoadIdentity();
			glTranslatef(x, y,0);
			switch(board[idx]){
				case kBlockEmpty:
					glColor4f(0.1f, 0.1f, 0.1f, 1);			
					break;
				case kBlockRock:
					glColor4f(0.5f, 0.5f, 0.5f, 1);			
					break;
				case kBlockFood:
					glColor4f(1.0f, 0.5f, 0.0f, 1);			
					break;
				case kBlockSnake:
					glColor4f(0.0f, 1.0f, 0.0f, 1);			
					break;
				default:
					glColor4f(1.0f, 0.0f, 0.0f, 1);			
					break;
			}
			
			// draw block
			glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
		}
	}

    // This application only creates a single color renderbuffer which is already bound at this point.
    // This call is redundant, but needed if dealing with multiple renderbuffers.
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER_OES];
}

- (BOOL)resizeFromLayer:(CAEAGLLayer *)layer
{	
    // Allocate color buffer backing based on the current layer size
    glBindRenderbufferOES(GL_RENDERBUFFER_OES, colorRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER_OES fromDrawable:layer];
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_WIDTH_OES, &backingWidth);
    glGetRenderbufferParameterivOES(GL_RENDERBUFFER_OES, GL_RENDERBUFFER_HEIGHT_OES, &backingHeight);

    if (glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES) != GL_FRAMEBUFFER_COMPLETE_OES)
    {
        NSLog(@"Failed to make complete framebuffer object %x", glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES));
        return NO;
    }

    return YES;
}

- (void)dealloc
{
    // Tear down GL
    if (defaultFramebuffer)
    {
        glDeleteFramebuffersOES(1, &defaultFramebuffer);
        defaultFramebuffer = 0;
    }

    if (colorRenderbuffer)
    {
        glDeleteRenderbuffersOES(1, &colorRenderbuffer);
        colorRenderbuffer = 0;
    }

    // Tear down context
    if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];

    [context release];
    context = nil;

    [super dealloc];
}

@end
