//
//  ES2Renderer.m
//  Snake
//
//  Created by Vahid Kazemi on 7/12/10.
//  Copyright KTH 2010. All rights reserved.
//

#import "ES2Renderer.h"
#import "World.h"

// uniform index
enum {
    UNIFORM_TRANSLATE,
	UNIFORM_COLOR,
	UNIFORM_TEXTURE,
    NUM_UNIFORMS
};
GLint uniforms[NUM_UNIFORMS];

// attribute index
enum {
    ATTRIB_VERTEX,
	ATTRIB_TEXCOORD,
    NUM_ATTRIBUTES
};

@interface ES2Renderer (PrivateMethods)
- (BOOL)loadShaders;
- (BOOL)compileShader:(GLuint *)shader type:(GLenum)type file:(NSString *)file;
- (BOOL)linkProgram:(GLuint)prog;
- (BOOL)validateProgram:(GLuint)prog;
@end

@implementation ES2Renderer

// Create an OpenGL ES 2.0 context
- (id)init
{
    if ((self = [super init]))
    {		
        context = [[EAGLContext alloc] initWithAPI:kEAGLRenderingAPIOpenGLES2];

        if (!context || ![EAGLContext setCurrentContext:context] || ![self loadShaders])
        {
            [self release];
            return nil;
        }
		
        // Create default framebuffer object. The backing will be allocated for the current layer in -resizeFromLayer
        glGenFramebuffers(1, &defaultFramebuffer);
        glGenRenderbuffers(1, &colorRenderbuffer);
        glBindFramebuffer(GL_FRAMEBUFFER, defaultFramebuffer);
        glBindRenderbuffer(GL_RENDERBUFFER, colorRenderbuffer);
        glFramebufferRenderbuffer(GL_FRAMEBUFFER, GL_COLOR_ATTACHMENT0, GL_RENDERBUFFER, colorRenderbuffer);
		
		texture = [self loadTexture:@"block.png"];
		accel.x = accel.y = accel.z = 1;
    }

    return self;
}
- (GLuint)loadTexture:(NSString *)filename
{
	UIImage *image = [UIImage imageNamed:filename];
    if (image == nil){
        NSLog(@"Do real error checking here");
		return 0;
	}
	
    GLuint width = CGImageGetWidth(image.CGImage);
    GLuint height = CGImageGetHeight(image.CGImage);
    void *imageData = malloc( height * width * 4 );
	CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
	CGContextRef bitmapContext = CGBitmapContextCreate( imageData, width, height, 8, 4 * width, colorSpace, kCGImageAlphaPremultipliedLast | kCGBitmapByteOrder32Big );
    CGColorSpaceRelease( colorSpace );
    CGContextClearRect( bitmapContext, CGRectMake( 0, 0, width, height ) );
    CGContextTranslateCTM( bitmapContext, 0, height - height );
    CGContextDrawImage( bitmapContext, CGRectMake( 0, 0, width, height ), image.CGImage );
	
	GLuint texid;
	glGenTextures(1, &texid);
	glBindTexture(GL_TEXTURE_2D, texid);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, width, height, 0, GL_RGBA, GL_UNSIGNED_BYTE, imageData);
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR); 
    glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);

    CGContextRelease(bitmapContext);
	
    free(imageData);
	
	return texid;
}

- (void)accelerate:(UIAcceleration*)acc 
{
	if(fabsf(acc.x)>1.5f || fabsf(acc.y)>1.5f || fabsf(acc.z)>1.5){
		accel.x = 10*acc.x;
		accel.y = 10*acc.y;
		accel.z = 10*acc.z;
	}
	accel.x = accel.x*0.9 + 0.1;
	accel.y = accel.y*0.9 + 0.1;
	accel.z = accel.z*0.9 + 0.1;
}

- (void)render
{
    // Replace the implementation of this method to do your own custom drawing
	
	float blockWidth = 2.0f/kBoardWidth;
	float blockHeight = 2.0f/kBoardHeight;

    GLfloat squareVertices[] = {
        0,			0,
        blockWidth,	0,
        0,			blockHeight,
        blockWidth,	blockHeight
    };
	
	GLfloat texcoords[] = {
		0, 0,
		1, 0,
		0, 1,
		1, 1,
	};

    // This application only creates a single context which is already set current at this point.
    // This call is redundant, but needed if dealing with multiple contexts.
    [EAGLContext setCurrentContext:context];

    // This application only creates a single default framebuffer which is already bound at this point.
    // This call is redundant, but needed if dealing with multiple framebuffers.
    glBindFramebuffer(GL_FRAMEBUFFER, defaultFramebuffer);
    glViewport(0, 0, backingWidth, backingHeight);

	static int lastColorCount = 0;
	static float lastColor[] = {0, 0, 0};
	static float newColor[] = {0, 0, 0};
	if(lastColorCount--<=0){
		float sum = 0.0001f;
		for(int i=0; i<3; i++){
			newColor[i] = (float)rand()/RAND_MAX;
			sum += newColor[i];
		}
		float factor = 0.5f*(float)rand()/RAND_MAX/sum;
		for(int i=0; i<3; i++){
			newColor[i] *= factor;
		}
		lastColorCount = 100;
	}
	for(int i=0; i<3; i++){
		lastColor[i] = lastColor[i]*0.99f + newColor[i]*0.01f;
	}
	
    glClearColor(lastColor[0], lastColor[1], lastColor[2], 1.0f);
    glClear(GL_COLOR_BUFFER_BIT);

    // Use shader program
    glUseProgram(program);

    // Update attribute values
    glVertexAttribPointer(ATTRIB_VERTEX, 2, GL_FLOAT, 0, 0, squareVertices);
    glEnableVertexAttribArray(ATTRIB_VERTEX);
    glVertexAttribPointer(ATTRIB_TEXCOORD, 2, GL_FLOAT, 0, 0, texcoords);
    glEnableVertexAttribArray(ATTRIB_TEXCOORD);

    // Validate program before drawing. This is a good check, but only really necessary in a debug build.
    // DEBUG macro must be defined in your debug configurations if that's not already the case.
#if defined(DEBUG)
    if (![self validateProgram:program])
    {
        NSLog(@"Failed to validate program: %d", program);
        return;
    }
#endif
	
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE);
		
	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, texture);
	glUniform1i(uniforms[UNIFORM_TEXTURE], 0);

	static Vector2f lastRandom[kBoardWidth*kBoardHeight] = {0};
	static Vector2f newRandom[kBoardWidth*kBoardHeight] = {0};
	static int lastRandCount[kBoardWidth*kBoardHeight] = {0};
	
	static float time = 0;
	time+=1.0f/60;
	
	// Draw board
	BlockType* board = [[World getInstance] getBoard];
	for(int j=0; j<kBoardHeight; j++){
		for(int i=0; i<kBoardWidth; i++){	
			int idx = i+j*kBoardWidth;

			// create a random number		
			if(lastRandCount[idx]<=0){
				newRandom[idx].x = 0.02f*((float)rand()*2/RAND_MAX-1)*accel.x;
				newRandom[idx].y = 0.02f*((float)rand()*2/RAND_MAX-1)*accel.y;
				lastRandCount[idx] = 20;
			}
			else{
				lastRandCount[idx]--;
			}
			lastRandom[idx].x = newRandom[idx].x*0.1f+lastRandom[idx].x*0.9f;
			lastRandom[idx].y = newRandom[idx].y*0.1f+lastRandom[idx].y*0.9f;

			float x = i*blockWidth-1+lastRandom[idx].x;
			float y = 1.0f-(j+1)*blockHeight+lastRandom[idx].y;
			glUniform2f(uniforms[UNIFORM_TRANSLATE], x, y);
			switch(board[idx]){
				case kBlockEmpty:
					glUniform4f(uniforms[UNIFORM_COLOR], 0.1f, 0.1f, 0.1f, 1);			
					break;
				case kBlockRock:
					glUniform4f(uniforms[UNIFORM_COLOR], 0.5f, 0.5f, 0.5f, 1);			
					break;
				case kBlockFood:
				{
					float effect = sinf(time*10)*0.3f + 0.7f;
					glUniform4f(uniforms[UNIFORM_COLOR], 1.0f*effect, 0.5f*effect, 0.0f, 1);			
					break;
				}
				case kBlockSnake:
					glUniform4f(uniforms[UNIFORM_COLOR], 0.0f, 1.0f, 0.0f, 1);			
					break;
				default:
					glUniform4f(uniforms[UNIFORM_COLOR], 1.0f, 0.0f, 0.0f, 1);			
					break;
			}
			
			// draw block
			glDrawArrays(GL_TRIANGLE_STRIP, 0, 4);
		}
	}

    // This application only creates a single color renderbuffer which is already bound at this point.
    // This call is redundant, but needed if dealing with multiple renderbuffers.
    glBindRenderbuffer(GL_RENDERBUFFER, colorRenderbuffer);
    [context presentRenderbuffer:GL_RENDERBUFFER];
}

- (BOOL)compileShader:(GLuint *)shader type:(GLenum)type file:(NSString *)file
{
    GLint status;
    const GLchar *source;

    source = (GLchar *)[[NSString stringWithContentsOfFile:file encoding:NSUTF8StringEncoding error:nil] UTF8String];
    if (!source)
    {
        NSLog(@"Failed to load vertex shader");
        return FALSE;
    }

    *shader = glCreateShader(type);
    glShaderSource(*shader, 1, &source, NULL);
    glCompileShader(*shader);

#if defined(DEBUG)
    GLint logLength;
    glGetShaderiv(*shader, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetShaderInfoLog(*shader, logLength, &logLength, log);
        NSLog(@"Shader compile log:\n%s", log);
        free(log);
    }
#endif

    glGetShaderiv(*shader, GL_COMPILE_STATUS, &status);
    if (status == 0)
    {
        glDeleteShader(*shader);
        return FALSE;
    }

    return TRUE;
}

- (BOOL)linkProgram:(GLuint)prog
{
    GLint status;

    glLinkProgram(prog);

#if defined(DEBUG)
    GLint logLength;
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetProgramInfoLog(prog, logLength, &logLength, log);
        NSLog(@"Program link log:\n%s", log);
        free(log);
    }
#endif

    glGetProgramiv(prog, GL_LINK_STATUS, &status);
    if (status == 0)
        return FALSE;

    return TRUE;
}

- (BOOL)validateProgram:(GLuint)prog
{
    GLint logLength, status;

    glValidateProgram(prog);
    glGetProgramiv(prog, GL_INFO_LOG_LENGTH, &logLength);
    if (logLength > 0)
    {
        GLchar *log = (GLchar *)malloc(logLength);
        glGetProgramInfoLog(prog, logLength, &logLength, log);
        NSLog(@"Program validate log:\n%s", log);
        free(log);
    }

    glGetProgramiv(prog, GL_VALIDATE_STATUS, &status);
    if (status == 0)
        return FALSE;

    return TRUE;
}

- (BOOL)loadShaders
{
    GLuint vertShader, fragShader;
    NSString *vertShaderPathname, *fragShaderPathname;

    // Create shader program
    program = glCreateProgram();

    // Create and compile vertex shader
    vertShaderPathname = [[NSBundle mainBundle] pathForResource:@"Shader" ofType:@"vsh"];
    if (![self compileShader:&vertShader type:GL_VERTEX_SHADER file:vertShaderPathname])
    {
        NSLog(@"Failed to compile vertex shader");
        return FALSE;
    }

    // Create and compile fragment shader
    fragShaderPathname = [[NSBundle mainBundle] pathForResource:@"Shader" ofType:@"fsh"];
    if (![self compileShader:&fragShader type:GL_FRAGMENT_SHADER file:fragShaderPathname])
    {
        NSLog(@"Failed to compile fragment shader");
        return FALSE;
    }

    // Attach vertex shader to program
    glAttachShader(program, vertShader);

    // Attach fragment shader to program
    glAttachShader(program, fragShader);

    // Bind attribute locations
    // this needs to be done prior to linking
    glBindAttribLocation(program, ATTRIB_VERTEX, "position");
    // Link program
    if (![self linkProgram:program])
    {
        NSLog(@"Failed to link program: %d", program);

        if (vertShader)
        {
            glDeleteShader(vertShader);
            vertShader = 0;
        }
        if (fragShader)
        {
            glDeleteShader(fragShader);
            fragShader = 0;
        }
        if (program)
        {
            glDeleteProgram(program);
            program = 0;
        }
        
        return FALSE;
    }

    // Get uniform locations
    uniforms[UNIFORM_TRANSLATE] = glGetUniformLocation(program, "translate");
    uniforms[UNIFORM_COLOR] = glGetUniformLocation(program, "color");
	uniforms[UNIFORM_TEXTURE] = glGetUniformLocation(program, "texture0");

    // Release vertex and fragment shaders
    if (vertShader)
        glDeleteShader(vertShader);
    if (fragShader)
        glDeleteShader(fragShader);

    return TRUE;
}

- (BOOL)resizeFromLayer:(CAEAGLLayer *)layer
{
    // Allocate color buffer backing based on the current layer size
    glBindRenderbuffer(GL_RENDERBUFFER, colorRenderbuffer);
    [context renderbufferStorage:GL_RENDERBUFFER fromDrawable:layer];
    glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_WIDTH, &backingWidth);
    glGetRenderbufferParameteriv(GL_RENDERBUFFER, GL_RENDERBUFFER_HEIGHT, &backingHeight);

    if (glCheckFramebufferStatus(GL_FRAMEBUFFER) != GL_FRAMEBUFFER_COMPLETE)
    {
        NSLog(@"Failed to make complete framebuffer object %x", glCheckFramebufferStatus(GL_FRAMEBUFFER));
        return NO;
    }

    return YES;
}

- (void)dealloc
{
    // Tear down GL
    if (defaultFramebuffer)
    {
        glDeleteFramebuffers(1, &defaultFramebuffer);
        defaultFramebuffer = 0;
    }

    if (colorRenderbuffer)
    {
        glDeleteRenderbuffers(1, &colorRenderbuffer);
        colorRenderbuffer = 0;
    }

    if (program)
    {
        glDeleteProgram(program);
        program = 0;
    }

    // Tear down context
    if ([EAGLContext currentContext] == context)
        [EAGLContext setCurrentContext:nil];

    [context release];
    context = nil;

    [super dealloc];
}

@end
