//
//  World.m
//  Snake
//
//  Created by Vahid Kazemi on 7/13/10.
//  Copyright 2010 KTH. All rights reserved.
//

#import "World.h"

static World* theWorld = nil;

@implementation World

+ (World*) getInstance {
	@synchronized( self ) {
		if(theWorld==nil){
			theWorld = [[World alloc] init];
		}
	}
	return theWorld;
}

+ (void) releaseInstance {
	@synchronized( self ) {
		if(theWorld!=nil){
			[theWorld release];
			theWorld = nil;
		}
	}
}

+ (int) calculateRank:(int) points {
	static const int numRanks = 7;
	static const int rankPoints[] = {0, 200, 500, 900, 1400, 2000, 2700};
	int rank = 0;
	for(int i=0; i<numRanks; i++){
		if(points >= rankPoints[i]){
			rank = i;
		}
	}
	return rank;
}

+ (NSString*) getRankTitle : (int)rank {
	static NSString* rankTitles[] = {@"Snake", @"Viper", @"Elapid", @"Python", @"Boa", @"Dragon", @"Mother of God"};
	return rankTitles[rank];
}

+ (NSString*) getAchivementID:(int)rank {
	static NSString* achivementID[] = {nil, @"520904", @"520914", @"520924", @"520934", @"520944", @"520954"};
	return achivementID[rank];
}

- (int) getPoints {
	return points;
}

- (NSMutableArray*) getSnakes {
	return snakes;
}

- (Snake*) getPlayer {
	return player;
}

- (BlockType*) getBoard {
	return board;
}

- (bool) isPaused {
	return gamePaused;
}

- (bool) isStopped {
	return gameStoped;
}

- (SoundMgr*) getSoundMgr {
	return soundMgr;
}

- (id) init {
    if ((self = [super init])){
		soundMgr = [[SoundMgr alloc] init];
		snakes = [[NSMutableArray alloc] init];
		player = nil;
		playerDir = 0;
		gamePaused = true;
		gameStoped = true;
		points = 0;
		set2i(&food, 0, 0);
		memset(board, 0, sizeof(BlockType)*kBoardWidth*kBoardHeight);
		[NSTimer scheduledTimerWithTimeInterval:0.25f 
										 target:self 
									   selector:@selector(update) 
									   userInfo:nil 
										repeats:YES];
		[self start];
	}
	
	return self;
}

- (void) dealloc {
	[soundMgr release];
	soundMgr = nil;
	player = nil;
	
	// free snakes
	while(snakes.count>0){
		[snakes removeLastObject];
	}
	[snakes release];
	snakes = nil;
	
	[super dealloc];
}

- (void) start {
	// clear board
	points = 0;
	memset(board, 0, sizeof(BlockType)*kBoardWidth*kBoardHeight);
	while(snakes.count>0){
		[snakes removeLastObject];
	}
	
	// generate Snakes
	for(int i=0; i<3; ++i){
		int randX = rand()%(kBoardWidth/2-1);
		int randY = rand()%(kBoardHeight/3-1) + i*kBoardHeight/3;
		Snake* w = [[Snake alloc] initWithID:(i+kBlockSnake) 
									   Dir:kDirRight
									 fromX:randX andY:randY 
									   toX:randX+2 andY:randY];
		[snakes addObject:w];
		for(int j=0; j<w.blocks.count; j++){
			SnakeBlock* b = [w.blocks objectAtIndex:j];
			board[b.pos.x+b.pos.y*kBoardWidth] = (BlockType)w.no;
		}
        [w release];
	}
	player = [snakes objectAtIndex:0];
	playerDir = kDirRight;
	
	int numRocks = rand()%5;
	for(int i=0; i<numRocks; i++){
		int x = rand()%kBoardWidth;
		int y = rand()%kBoardHeight;
		if(board[x+y*kBoardWidth]==kBlockEmpty){
			board[x+y*kBoardWidth] = kBlockRock;
		}
	}
	
	food.x = rand()%(kBoardWidth/2-1)+(kBoardWidth/2);
	food.y = rand()%kBoardHeight;
	while(board[food.x+food.y*kBoardWidth]!=kBlockEmpty){
		food.x = rand()%(kBoardWidth/2-1)+(kBoardWidth/2);
		food.y = rand()%kBoardHeight;
	}
	board[food.x+food.y*kBoardWidth] = kBlockFood;
	
	gameStoped = false;
	gamePaused = true;
}

- (void) stop {
	gameStoped = true;
	player = nil;
}

- (void) toggleGameState {
	if(gameStoped){
		[self start];
	}
	else{
		if(gamePaused){
			gamePaused = false;
			[soundMgr play:kSoundBackground];
		}
		else{
			gamePaused = true;
			[soundMgr pause:kSoundBackground];
		}
	}
}

// input functions
- (void) move: (Snake*)snake {
	if(gamePaused)return;
	if(snake==nil)return;
	if(snake.blocks.count==0){
		NSLog(@"Error: Snake blocks count is zero.");
		return;
	}
	Vector2i pos = [[snake.blocks lastObject] pos];
	pos.x = (pos.x+snake.dir.x+kBoardWidth)%kBoardWidth;
	pos.y = (pos.y+snake.dir.y+kBoardHeight)%kBoardHeight;
	BlockType bt = board[pos.x+pos.y*kBoardWidth];
	switch (bt) {
		case kBlockEmpty:
        {
			if(snake.tail>0){
				snake.tail--;
				if(snake==player){
					[soundMgr play:kSoundMerge];
					points+=10;
				}
			}
			else{
				SnakeBlock* b = [snake.blocks objectAtIndex:0];
				board[b.pos.x+b.pos.y*kBoardWidth] = kBlockEmpty;
				[snake.blocks removeObjectAtIndex:0];
			}
            SnakeBlock* block = [[SnakeBlock alloc] initWithVector:pos];
			[snake.blocks addObject:block];
            [block release];
			board[pos.x+pos.y*kBoardWidth] = (BlockType)snake.no;
			break;
        }
		case kBlockRock:
        {
			snake.die = true;
			if(snake==player)
				[soundMgr play:kSoundCrash];
			break;
        }
		case kBlockFood:
        {
            SnakeBlock* block = [[SnakeBlock alloc] initWithVector:pos];
			[snake.blocks addObject:block];
            [block release];
			board[pos.x+pos.y*kBoardWidth] = (BlockType)snake.no;
			if(snake==player){
				[soundMgr play:kSoundEat];
				points+=25;
			}

			// generate new food
			while(board[food.x+food.y*kBoardWidth]!=kBlockEmpty){
				  food.x = rand()%kBoardWidth;
				  food.y = rand()%kBoardHeight;
			}
			board[food.x+food.y*kBoardWidth] = kBlockFood;
			break;
        }
		default:
        {
			snake.die = true;
			for(int i=0; i<snakes.count; i++){
				Snake* w = [snakes objectAtIndex:i];
				if(w.no==bt){
					w.tail += snake.blocks.count;
					break;
				}
			}
			if(snake==player)
				[soundMgr play:kSoundCrash];
			break;
        }
	}
}

- (bool) turn: (Snake*)snake To:(int)d {
	if(gamePaused)return false;
	if(snake==nil)return false;	   
	Vector2i dir = [Snake getDirByID:d];
	if((dir.x+snake.dir.x)*(dir.y+snake.dir.y)!=0){
		snake.dir = dir;
		return true;
	}
	return false;
}

- (void) control:(int)dir {
	playerDir = dir;
	[soundMgr play:kSoundRotate];
}

- (void) update {
	if( gamePaused )return;
	
	for(int i=0; i<snakes.count; i++){
		Snake* w = [snakes objectAtIndex:i];
		if(!w.die){
			// AI
			if(w!=player){
				Vector2i pos = [[w.blocks lastObject] pos];
				int bestRank = -0xfffff, bestDir=0;
				for(int k=0; k<4; k++){
					Vector2i dir = [Snake getDirByID:k];

					// calculate new position
					Vector2i npos = add2i(&pos, &dir);
					int idx = npos.x+npos.y*kBoardWidth;
					
					// continue if a block or Snake in front
					if( board[idx]==kBlockRock || board[idx]>=kBlockSnake )
						continue;
					
					// calculate move rank
					int rank = (ABS(food.x-pos.x)+ABS(food.y-pos.y))-
						(ABS(food.x-npos.x)+ABS(food.y-npos.y));
					
					if(rank >= bestRank){
						bestRank = rank;
						bestDir = k;
					}
				}
				[self turn:w To:bestDir];
			}
			// player control
			else{
				[self turn:w To:playerDir];
			}
			[self move:w];
		}
		else{
			if(w.blocks.count>0){
				SnakeBlock* b = [w.blocks objectAtIndex:0];
				board[b.pos.x+b.pos.y*kBoardWidth] = kBlockEmpty;
				[w.blocks removeObjectAtIndex:0];
				if(w==player){
					[soundMgr play:kSoundMerge];
				}
			}
			else{
				NSLog(@"Error: blocks count is zero.");
			}
			// remove Snake if size is zero
			if(w.blocks.count==0){
				if(w==player){
					player = nil;
				}
				[snakes removeObjectAtIndex:i];
				i--;
			}
		}
	}
	
	if((player==nil || player.die) && !gameStoped){
		[self stop];
	}
}

@end
